# Test converting 32-bit integer images using the pawt package.
#
# Copyright: 2021-2022 Paul Obermeier (obermeier@poSoft.de)
# Distributed under BSD license.

if { [file exists "TestUtils.tcl"] } {
    source "TestUtils.tcl"
}

package require pawt

if { ! [pawt::HavePkg "Img"] } {
    puts "Error: pawt test suite needs the Img extension."
    exit 1
}

if { [package vcompare [package version "Img"] 1.4.14] < 0 } {
    puts "Error: pawt test suite for int values needs Img > 1.4.13."
    exit 1
}

if { ! [file isdirectory "testOut"] } {
    file mkdir "testOut"
}

set testCase "03_Convert-int"

set inFileList { "int-1chan-bu" "int-1chan-td" "int-3chan-bu" "int-3chan-td" }

foreach inFilePrefix $inFileList {
    set inFileName [format "%s.%s" $inFilePrefix "raw"]
    set inFilePath [file join "testIn" $inFileName]
    puts "Reading file $inFilePath ..."
    set rawDict [pawt::raw::ReadImageFile $inFilePath]

    set rawImgAsListTD [pawt GetImageAsList   rawDict -scanorder "TopDown"]
    set rawImgAsListBU [pawt GetImageAsList   rawDict -scanorder "BottomUp"]
    set rawImgAsMatTD  [pawt GetImageAsMatrix rawDict -scanorder "TopDown"]
    set rawImgAsMatBU  [pawt GetImageAsMatrix rawDict -scanorder "BottomUp"]

    set outRawName [format "%s_%s.%s" $testCase $inFilePrefix "raw"]
    set outRawPath [file join "testOut" $outRawName]
    puts "Writing file $outRawPath ..."
    pawt::raw::WriteImageFile $rawDict $outRawPath

    pawtTest CheckFile $inFilePath $outRawPath "Check $inFilePath vs. $outRawPath"


    set outFitsName [format "%s_%s.%s" $testCase $inFilePrefix "fits"]
    set outFitsPath [file join "testOut" $outFitsName]
    puts "Writing file $outFitsPath ..."
    pawt::fits::WriteImageFile $rawDict $outFitsPath

    set fitsDict [pawt::fits::ReadImageFile $outFitsPath]
    set fitsImgAsListTD [pawt GetImageAsList   fitsDict -scanorder "TopDown"]
    set fitsImgAsListBU [pawt GetImageAsList   fitsDict -scanorder "BottomUp"]
    set fitsImgAsMatTD  [pawt GetImageAsMatrix fitsDict -scanorder "TopDown"]
    set fitsImgAsMatBU  [pawt GetImageAsMatrix fitsDict -scanorder "BottomUp"]

    pawtTest CheckList   $rawImgAsListTD $fitsImgAsListTD "Check RAW vs. FITS TD list"
    pawtTest CheckList   $rawImgAsListBU $fitsImgAsListBU "Check RAW vs. FITS BU list"
    pawtTest CheckMatrix $rawImgAsMatTD  $fitsImgAsMatTD  "Check RAW vs. FITS TD matrix"
    pawtTest CheckMatrix $rawImgAsMatBU  $fitsImgAsMatBU  "Check RAW vs. FITS BU matrix"

   
    if { [string match "*3chan*" $inFilePrefix] } {
        set flirDict [pawtTest CreateDictUnsupported [pawt GetImageWidth rawDict] [pawt GetImageHeight rawDict]]
        set outFlirName "3-channel int FLIR images not supported"
    } else {
        set outFlirName [format "%s_%s.%s" $testCase $inFilePrefix "fpf"]
        set outFlirPath [file join "testOut" $outFlirName]
        puts "Writing file $outFlirPath ..."
        pawt::flir::WriteImageFile $rawDict $outFlirPath

        set flirDict [pawt::flir::ReadImageFile $outFlirPath]
        set flirImgAsListTD [pawt GetImageAsList   flirDict -scanorder "TopDown"]
        set flirImgAsListBU [pawt GetImageAsList   flirDict -scanorder "BottomUp"]
        set flirImgAsMatTD  [pawt GetImageAsMatrix flirDict -scanorder "TopDown"]
        set flirImgAsMatBU  [pawt GetImageAsMatrix flirDict -scanorder "BottomUp"]

        pawtTest CheckList   $rawImgAsListTD $flirImgAsListTD "Check RAW vs. FLIR TD list"
        pawtTest CheckList   $rawImgAsListBU $flirImgAsListBU "Check RAW vs. FLIR BU list"
        pawtTest CheckMatrix $rawImgAsMatTD  $flirImgAsMatTD  "Check RAW vs. FLIR TD matrix"
        pawtTest CheckMatrix $rawImgAsMatBU  $flirImgAsMatBU  "Check RAW vs. FLIR BU matrix"
    }

    set pgmbDict [pawtTest CreateDictUnsupported [pawt GetImageWidth rawDict] [pawt GetImageHeight rawDict]]
    set outPgmbName "PPM int binary images not supported"

    set pgmaDict [pawtTest CreateDictUnsupported [pawt GetImageWidth rawDict] [pawt GetImageHeight rawDict]]
    set outPgmaName "PPM int ASCII images not supported"

    pawtTest AddImgDicts $rawDict  $outRawName \
                         $fitsDict $outFitsName \
                         $flirDict $outFlirName \
                         $pgmbDict $outPgmbName \
                         $pgmaDict $outPgmaName \
                         -title $testCase -zoom 50
    puts ""
}


if { [lindex $argv 0] eq "auto" } {
    update
    after 500
    exit 0
}
